// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of suplibxx
//
// suplibxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// suplibxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <algorithm>
#include <set>
#include <pcre.h>
#include <Exception/Exception.h>

#include "colselect.h"

namespace suplib {
  /*!
   * \ingroup colselect
   * \brief select columns based on exact/regex matching/exclusion.
   *
   *
   * \param icolumns vector of column names on which to operate
   * \param exact_add column names to add to output
   * \param regex_add regular expressions used to add to output
   * \param exact_del column names to exclude from output
   * \param regex_del regular expressions used to exclude from output
   * \param ocolumns set of column names which were selected
   *
   * \exception Exception errors related to pattern matching
   *
   * colselect places strings from the input vector in the output
   * vector based on matches specified by the parameters.  Strings
   * appearing in the exact_del parameter are excluded from the output
   * vector.  Strings which match the regular expression in regex_del
   * are excluded from the output vector.  Strings which do not match
   * any of the *_del parameters but are either present in exact_add
   * or match a regular expression in regex_add are added to the
   * output set.
   *
   */
void colselect( 
  const std::vector<std::string>& icolumns,
  const std::vector<std::string>& exact_add, 
  const std::vector<std::string>& regex_add,
  const std::vector<std::string>& exact_del,
  const std::vector<std::string>& regex_del,
  std::vector<std::string>& ocolumns
  ) {
  
  if ( 0 == (exact_add.size( ) + regex_add.size( ) + 
	     exact_del.size( ) + regex_del.size( )) ) {
    ocolumns = icolumns;

  } else {
    std::set<std::string> 
      add_set, exact_add_set, regex_add_set, 
      del_set, exact_del_set, regex_del_set, 
      column_set;

    exact_add_set.insert( exact_add.begin( ), exact_add.end( ) );
    regex_add_set.insert( regex_add.begin( ), regex_add.end( ) );
    exact_del_set.insert( exact_del.begin( ), exact_del.end( ) );
    regex_del_set.insert( regex_del.begin( ), regex_del.end( ) );
    column_set.insert(     icolumns.begin( ),  icolumns.end( ) );

    ocolumns.clear();

    // Add exact matches...
    set_intersection( column_set.begin( ), column_set.end( ),
		      exact_add_set.begin( ), exact_add_set.end( ),
		      inserter( add_set, add_set.begin( ) )
		      );

    // Add regex matches...
    for ( const auto& pat : regex_add_set )
      for ( auto& nam : column_set )
	if ( match( nam, pat ) )
	  add_set.insert( nam );

    // Remove exact matches...
    set_difference( column_set.begin( ), column_set.end( ),
		    exact_del_set.begin( ), exact_del_set.end( ),
		    inserter( del_set, del_set.begin( ) )
		    );
    
    // Remove regex matches...
    for ( auto& pat : regex_del_set )
	 
        for ( auto nam = del_set.begin( ); nam != del_set.end( ) ; ) { 
            if ( match( *nam, pat ) )
                nam = del_set.erase( nam );
            else
                ++nam;
        }

    if ( exact_add.size( ) + regex_add.size( ) ) {
      std::set<std::string> ocolumn_set;
      set_intersection( del_set.begin( ), del_set.end( ),
			add_set.begin( ), add_set.end( ),
			inserter( ocolumn_set, ocolumn_set.begin( ) ) );
      
      for ( std::vector<std::string>::size_type idx = 0; idx < icolumns.size( ); idx++ ) {
	if ( ocolumn_set.end( ) != ocolumn_set.find( icolumns[idx] ) ) {
	  ocolumns.push_back( icolumns[idx] );
	  
	}
      }
    } else {
      for ( std::vector<std::string>::size_type idx = 0; idx != icolumns.size( ); idx++ ) {
	if ( del_set.end( ) != del_set.find( icolumns[idx] ) ) {
	  ocolumns.push_back( icolumns[idx] );

	}
      }
    }
  }
}

} // namespace suplib
