// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of suplibxx
//
// suplibxx is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// suplibxx is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <fstream>
#include <string>
#include <cstdlib>
#include <algorithm>

#include <boost/filesystem.hpp>
namespace fs = boost::filesystem;


#include "io.h"

#define CATCH_CONFIG_MAIN
#include <catch2/catch.hpp>

namespace {
  const std::string TESTSTR1(  "  now, is, the, time  " );
  const std::string TESTSTR2(  "x" );
  const std::string TESTSTR3(  "for");
  const std::string TESTSTR4(  "FOR" );
  const std::string TESTSTR5(  "now, is, the, time  " );
  const std::string TESTSTR6(  "  now, is, the, time" );
  const std::string TESTSTR7(  "now, is, the, time" );
  const std::string TESTSTR10( "now,is,the,time,\"for,all,good\", men, to come" );
  const std::string TESTSTR11( "now" );
  const std::string TESTSTR12( " to come" );
  const std::string TESTSTR13( "# This is a comment" );
  const std::string TESTSTR14( "                            " );
  const std::string TESTSTR15( "/data/hxds3/pipe/data/950504/lin409y0z0.pha" );
  const std::string TESTSTR16( "10.3" );
  const std::string TESTSTR17( "10.3e4" );
  const std::string TESTSTR18( "1234" );
  const std::string TESTSTR19( "123456789123412345678912341234567891234" );
  const std::string TESTSTR20( "/data/hxds3/pipe/data/950504/lin409y0z0.pha\\" );
  const std::string TESTSTR21( "/data/hxds3/pipe`/data/950504/lin409y0z0.pha" );
}

struct TempFile {

  fs::path path;

  TempFile (const std::string& text ) {

    std::string tmpfile_template = (fs::temp_directory_path() / "suplibxx-XXXXXXXXXX" ).native();

    std::vector<char> tmpfile_char(tmpfile_template.length() + 1, 0);
    std::copy( tmpfile_template.begin(),  tmpfile_template.end(), tmpfile_char.begin() );

    close( mkstemp( &tmpfile_char.front() ) );

    path = &tmpfile_char.front();

    std::ofstream ofs( path.native() );
    ofs << text;
    ofs.close( );
  }

  operator std::string () const { return path.native(); }

  ~TempFile () {
    fs::remove( path );
  }

};


TEST_CASE( "physical", "[getrecord]" )
{
  std::string expect = TESTSTR1;
  std::string got;
  auto tmpfile = TempFile( expect );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_PHYS );
  REQUIRE( got == expect );
}

TEST_CASE( "physical | strip", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_PHYS | suplib::STRIP );
  REQUIRE( got == TESTSTR6 );
}

TEST_CASE( "logical", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "\\  \n" + TESTSTR1 + "\\  \n" + TESTSTR1 );
  std::string expect( TESTSTR1 + TESTSTR1 + TESTSTR1 );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL );
    
  REQUIRE( got == expect );
}


TEST_CASE( "logical | clean", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "\\  \n" + TESTSTR1 + "\\  \n" + TESTSTR1 );
  std::string expect(  TESTSTR1 + "\\\n" + TESTSTR1 + "\\\n" + TESTSTR1 );


  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::CLEAN );
  REQUIRE( got == expect );
}
    
TEST_CASE( "logical | strip, no continuation", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "\n" );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::STRIP );
  REQUIRE( got == TESTSTR6 );
}

TEST_CASE( "logical | strip, continuation", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "\\  \n" + TESTSTR1 + "\\  \n" + TESTSTR1 );
  std::string expect(  TESTSTR1 + TESTSTR1 + TESTSTR6 );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::STRIP);
  REQUIRE( got == expect );
}

TEST_CASE( "logical | strip | clean", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "\\  \n" + TESTSTR1 + "\\  \n" + TESTSTR1 );
  std::string expect(  TESTSTR1 + "\\\n" + TESTSTR1 + "\\\n" + TESTSTR6 );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::STRIP | suplib::CLEAN );
  REQUIRE( got == expect );
}

TEST_CASE( "logical, '-' is continuation character", "[getrecord]" )
{
  std::string got;
  auto tmpfile = TempFile( TESTSTR1 + "-  \n" + TESTSTR1 + "-  \n" + TESTSTR1 );
  std::string expect(  TESTSTR1 + TESTSTR1 + TESTSTR1 );

  std::ifstream ifs( tmpfile );
  suplib::getrecord( ifs, got, suplib::READ_LOGICAL, '\n', '-' );
  REQUIRE( got == expect );
}

  
namespace {
  const std::string TESTSTR1_0( "#This file contains an erroneous lower limit for the integer" );
  const std::string TESTSTR1_1( "junk,i,a,1,3,4,\"Level of debugging messages to be displayed " );
  const std::string TESTSTR1_2( "now is the time for all\"" );
  const std::string TESTSTR1_3( "verbose,i,a,1,-999999999999999999999999999,4,\"Level of debugging messages to be displayed\"" );
  const std::string TESTSTR1_4( "mode,s,h,\"q\",,,\"mode\"" );
}


struct MultiLine {

  TempFile tmpfile;

  MultiLine () :
    tmpfile( TempFile(
                      TESTSTR1_0 + "  \n" +
                      TESTSTR1_1 + "\\" + "  \n" +
                      TESTSTR1_2 + "  \n" +
                      TESTSTR1_3 + "  \n" +
                      TESTSTR1_4 + "  \n"
                      ) )
  {  }
};


TEST_CASE_METHOD( MultiLine, "logical | strip", "[getrecord]" )
{
  std::string got;
  std::ifstream ifs( tmpfile );
  int count = 0;

  while (suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::STRIP  )) {
    switch( count ) {
    case 0:
      REQUIRE( got == TESTSTR1_0 );
      break;
    case 1:
      REQUIRE( got == TESTSTR1_1 + TESTSTR1_2 );
      break;
    case 2:
      REQUIRE( got == TESTSTR1_3 );
      break;
    case 3:
      REQUIRE( got == TESTSTR1_4 );
      break;

    default:
      FAIL( "read too many lines" );
    }
    count++;
  }
}


TEST_CASE_METHOD( MultiLine, "logical | raw", "[getrecord]" )
{
  std::string got;
  std::ifstream ifs( tmpfile );
  int count = 0;

  while (suplib::getrecord( ifs, got, suplib::READ_LOGICAL | suplib::RAW  )) {
    switch( count ) {
    case 0:
      REQUIRE( got == TESTSTR1_0 + "  \n" );
      break;
    case 1:
      REQUIRE( got ==
               TESTSTR1_1
               +  "\\" 
               + "  \n"
               + TESTSTR1_2
               + "  \n"
               );
      break;
    case 2:
      REQUIRE( got == TESTSTR1_3 + "  \n" );
      break;
    case 3:
      REQUIRE( got == TESTSTR1_4 + "  \n" );
      break;
    default:
      FAIL( "read too many lines" );
    }
    count++;
  }
}

